# Copyright (c) 2020 by Marek Wydmuch
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import requests
import zipfile
import warnings
import re
from sys import stdout
from os import makedirs, path, remove
from sklearn.datasets import load_svmlight_file


# List of all available datasets
DATASETS = {
    'eurlex-4k': {
        'name': 'Eurlex-4K',
        'formats': ['tf-idf'],
        'subsets': ['train', 'test'],
        'tf-idf': {
            'url': 'https://drive.google.com/uc?export=download&id=0B3lPMIHmG6vGU0VTR1pCejFpWjg',
            'train': 'Eurlex/eurlex_train.txt',
            'test': 'Eurlex/eurlex_test.txt',
            'file_format': 'libsvm',
            'load_params': {'multilabel': True, 'zero_based': True, 'n_features': 5000, 'offset': 1}
        }
    },
    'amazoncat-13k': {
        'name': 'AmazonCat-13K',
        'formats': ['tf-idf'],
        'subsets': ['train', 'test'],
        'tf-idf': {
            'url': 'https://drive.google.com/uc?export=download&id=0B3lPMIHmG6vGa2tMbVJGdDNSMGc',
            'train': 'AmazonCat/amazonCat_train.txt',
            'test': 'AmazonCat/amazonCat_test.txt',
            'file_format': 'libsvm',
            'load_params': {'multilabel': True, 'zero_based': True, 'n_features': 203882, 'offset': 1}
        }
    },
    'amazoncat-14k': {
        'name': 'AmazonCat-14K',
        'formats': ['tf-idf'],
        'subsets': ['train', 'test'],
        'tf-idf': {
            'url': 'https://drive.google.com/uc?export=download&id=0B3lPMIHmG6vGaDFqU2E5U0dxS00',
            'train': 'AmazonCat-14K/amazonCat-14K_train.txt',
            'test': 'AmazonCat-14K/amazonCat-14K_test.txt',
            'file_format': 'libsvm',
            'load_params': {'multilabel': True, 'zero_based': True, 'n_features': 597540, 'offset': 1}
        }
    },
    'wiki10-31k': {
        'name': 'Wiki10-31K',
        'formats': ['tf-idf'],
        'subsets': ['train', 'test'],
        'tf-idf': {
            'url': 'https://drive.google.com/uc?export=download&id=0B3lPMIHmG6vGaDdOeGliWF9EOTA',
            'train': 'Wiki10/wiki10_train.txt',
            'test': 'Wiki10/wiki10_test.txt',
            'file_format': 'libsvm',
            'load_params': {'multilabel': True, 'zero_based': True, 'n_features': 101938, 'offset': 1}
        }
    },
    'deliciouslarge-200k': {
        'name': 'DeliciousLarge-200K',
        'formats': ['tf-idf'],
        'subsets': ['train', 'test'],
        'tf-idf': {
            'url': 'https://drive.google.com/uc?export=download&id=0B3lPMIHmG6vGR3lBWWYyVlhDLWM',
            'train': 'DeliciousLarge/deliciousLarge_train.txt',
            'test': 'DeliciousLarge/deliciousLarge_test.txt',
            'file_format': 'libsvm',
            'load_params': {'multilabel': True, 'zero_based': True, 'n_features': 782585, 'offset': 1}
        }
    },
    'wikilshtc-325k': {
        'name': 'WikiLSHTC-325K',
        'formats': ['tf-idf'],
        'subsets': ['train', 'test'],
        'tf-idf': {
            'url': 'https://drive.google.com/uc?export=download&id=0B3lPMIHmG6vGSHE1SWx4TVRva3c',
            'train': 'WikiLSHTC/wikiLSHTC_train.txt',
            'test': 'WikiLSHTC/wikiLSHTC_test.txt',
            'file_format': 'libsvm',
            'load_params': {'multilabel': True, 'zero_based': True, 'n_features': 1617899, 'offset': 1}
        }
    },
    'wikipedialarge-500k': {
        'name': 'WikipediaLarge-500K',
        'formats': ['tf-idf'],
        'subsets': ['train', 'test'],
        'tf-idf': {
            'url': 'https://drive.google.com/uc?export=download&id=0B3lPMIHmG6vGRmEzVDVkNjBMR3c',
            'train': 'WikipediaLarge-500K/WikipediaLarge-500K_train.txt',
            'test': 'WikipediaLarge-500K/WikipediaLarge-500K_test.txt',
            'file_format': 'libsvm',
            'load_params': {'multilabel': True, 'zero_based': True, 'n_features': 2381304, 'offset': 1}
        }
    },
    'amazon-670k': {
        'name': 'Amazon-670K',
        'formats': ['tf-idf'],
        'subsets': ['train', 'test'],
        'tf-idf': {
            'url': 'https://drive.google.com/uc?export=download&id=0B3lPMIHmG6vGdUJwRzltS1dvUVk',
            'train': 'Amazon/amazon_train.txt',
            'test': 'Amazon/amazon_test.txt',
            'file_format': 'libsvm',
            'load_params': {'multilabel': True, 'zero_based': True, 'n_features': 135909, 'offset': 1}
        }
    },
    'amazon-3m': {
        'name': 'Amazon-3M',
        'formats': ['tf-idf'],
        'subsets': ['train', 'test'],
        'tf-idf': {
            'url': 'https://drive.google.com/uc?export=download&id=0B3lPMIHmG6vGUEd4eTRxaWl3YkE',
            'train': 'Amazon-3M/amazon-3M_train.txt',
            'test': 'Amazon-3M/amazon-3M_test.txt',
            'file_format': 'libsvm',
            'load_params': {'multilabel': True, 'zero_based': True, 'n_features': 337067, 'offset': 1}
        }
    },
}

# Create datasets aliases
DATASETS['wiki10'] = DATASETS['wiki10-31k']
DATASETS['deliciouslarge'] = DATASETS['deliciouslarge-200k']
DATASETS['wikilshtc'] = DATASETS['wikilshtc-325k']
DATASETS['wikipedialarge'] = DATASETS['wikipedialarge-500k']


# Main function for downloading and loading datasets
def load_dataset(dataset, subset='train', format='tf-idf', root='./data', verbose=False):
    """
    Downloads the dataset from the internet and puts it in root directory.
    If dataset is already downloaded, it is not downloaded again.
    Then loads requested datasets into features matrix and labels.

    :param dataset: name of the dataset to load, case insensitive
    {'Eurlex-4K',
     'AmazonCat-13K',
     'AmazonCat-14K',
     'Wiki10-31K',
     'DeliciousLarge-200K',
     'WikiLSHTC-325K',
     'WikipediaLarge-500K',
     'Amazon-670K',
     'Amazon-3M'}
    :type dataset: str
    :param subset: subset of dataset to load into features matrix and labels {'train', 'test', 'all'}, defaults to 'train'
    :type subset: str, optional
    :param format: format of dataset to load {'tf-idf'}, defaults to 'tf-idf'
    :type format: str, optional
    :param root: location of datasets directory, defaults to ./data
    :type root: str, optional
    :param verbose: if True print downloading and loading progress, defaults to False
    :type verbose: bool, optional
    :return: (csr_matrix, list[list[int]]), tuple of features matrix and labels
    """
    dataset = dataset.lower()
    dataset_meta = DATASETS.get(dataset, None)

    if dataset_meta is None:
        raise ValueError("Dataset {} is not avialable".format(dataset))

    if format not in dataset_meta['formats']:
        raise ValueError("Format {} is not available for dataset {}".format(format, dataset))

    if subset is not None and subset not in dataset_meta['subsets']:
        raise ValueError("Subset {} is not available for dataset {}".format(format, dataset))

    dataset_dest = path.join(root, dataset_meta['name'] + '_' + format + ".zip")
    format_meta = dataset_meta[format]
    train_path = path.join(root, format_meta['train'])
    test_path = path.join(root, format_meta['test'])

    if not path.exists(train_path) or not path.exists(test_path):
        if 'drive.google.com' in format_meta['url']:
            GoogleDriveDownloader.download_file(format_meta['url'], dataset_dest, unzip=True, overwrite=True, delete_zip=True, verbose=verbose)

    if subset == 'train':
        return _load_file(train_path, format_meta['file_format'], format_meta['load_params'])
    elif subset == 'test':
        return _load_file(test_path, format_meta['file_format'], format_meta['load_params'])


# Helpers
def _load_file(filepath, format, load_params):
    if format == 'libsvm':
        return load_svmlight_file(filepath, **load_params)


# Based on: https://stackoverflow.com/questions/25010369/wget-curl-large-file-from-google-drive/39225039#39225039
class GoogleDriveDownloader:
    """
    Minimal class to download shared files from Google Drive.
    """

    CHUNK_SIZE = 32768
    DOWNLOAD_URL = 'https://drive.google.com/uc?export=download'

    @staticmethod
    def download_file(url, dest_path, overwrite=False, unzip=False, delete_zip=False, verbose=False):
        """
        Downloads a shared file from google drive into a given folder and optionally unzips it.
        :param url: file url to download
        :type url: str
        :param dest_path: the destination where to save the downloaded file
        :type dest_path: str
        :param overwrite: if True force redownload and overwrite, defaults to False
        :type overwrite: bool
        :param unzip: if True unzip a file, optional, defaults to False
        :type unzip: bool
        :param delete_zip: if True and unzips is True delete archive file after unziping, defaults to False
        :type delete_zip: bool
        :param verbose: if True print downloading progress, defaults to False
        :type verbose: bool
        :return: None
        """

        re_match = re.search('id=([\w\d]+)', url)
        file_id = re_match.group(1)

        destination_directory = path.dirname(dest_path)
        if not path.exists(destination_directory):
            makedirs(destination_directory)

        if not path.exists(dest_path) or overwrite:
            if verbose:
                print('Downloading {} into {} ... '.format(url, dest_path))

            session = requests.Session()
            response = session.get(GoogleDriveDownloader.DOWNLOAD_URL, params={'id': file_id}, stream=True)
            token = GoogleDriveDownloader._get_confirm_token(response)
            if token:
                params = {'id': file_id, 'confirm': token}
                response = session.get(GoogleDriveDownloader.DOWNLOAD_URL, params=params, stream=True)
            current_download_size = [0]
            GoogleDriveDownloader._save_response_content(response, dest_path, verbose, current_download_size)

            if unzip:
                try:
                    if verbose:
                        print('Unzipping ...')

                    with zipfile.ZipFile(dest_path, 'r') as z:
                        z.extractall(destination_directory)
                    if delete_zip:
                        remove(dest_path)

                except zipfile.BadZipfile:
                    warnings.warn('Ignoring `unzip` since "{}" does not look like a valid zip file'.format(file_id))

            if verbose:
                print('Done.')

    @staticmethod
    def _get_confirm_token(response):
        for key, value in response.cookies.items():
            if key.startswith('download_warning'):
                return value
        return None

    @staticmethod
    def _save_response_content(response, destination, verbose, current_size):
        with open(destination, 'wb') as f:
            for chunk in response.iter_content(GoogleDriveDownloader.CHUNK_SIZE):
                if chunk:  # filter out keep-alive new chunks
                    f.write(chunk)
                    if verbose:
                        print('\r' + GoogleDriveDownloader._sizeof_fmt(current_size[0]), end=' ')
                        stdout.flush()
                        current_size[0] += GoogleDriveDownloader.CHUNK_SIZE

    # From: https://stackoverflow.com/questions/1094841/reusable-library-to-get-human-readable-version-of-file-size
    @staticmethod
    def _sizeof_fmt(num, suffix='B'):
        for unit in ['', 'K', 'M', 'G', 'T', 'P', 'E', 'Z']:
            if abs(num) < 1024.0:
                return '{:.1f} {}{}'.format(num, unit, suffix)
            num /= 1024.0
        return '{:.1f} {}{}'.format(num, 'Y', suffix)



# Copyright (c) 2020 by Marek Wydmuch
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from math import log2


def _process_Y(Y):
    if not all((isinstance(y, list) or isinstance(y, tuple)) for y in Y):
        raise ValueError("Unsupported data format")

    return Y


def precision_at_k(Y_true, Y_pred, k=5):
    """
    Calculate precision at k

    :param Y_true: true labels
    :param Y_pred: ranking of predicted labels
    :param k: k
    :return: value of precision at k
    """
    Y_true = _process_Y(Y_true)
    Y_pred = _process_Y(Y_pred)

    sum = 0
    for t, p in zip(Y_true, Y_pred):
        p_at_k = 0
        for p_i in p[:k]:
            p_at_k += 1 if p_i in t else 0
        sum += p_at_k / k
    return sum / len(Y_true)


def recall_at_k(Y_true, Y_pred, k=5):
    """
    Calculate recall at k

    :param Y_true: true labels
    :param Y_pred: ranking of predicted labels
    :param k: k
    :return: value of recall at k
    """
    Y_true = _process_Y(Y_true)
    Y_pred = _process_Y(Y_pred)

    sum = 0
    for t, p in zip(Y_true, Y_pred):
        r_at_k = 0
        for p_i in p[:k]:
            r_at_k += 1 if p_i in t else 0
        sum += r_at_k / len(t)
    return sum / len(Y_true)



def coverage_at_k(Y_true, Y_pred, k=5):
    """
    Calculate coverage at k

    :param Y_true: true labels
    :param Y_pred: ranking of predicted labels
    :param k: k
    :return: value of coverage at k
    """

    Y_true = _process_Y(Y_true)
    Y_pred = _process_Y(Y_pred)

    uniq_t = set()
    uniq_tp = set()
    for t, p in zip(Y_true, Y_pred):
        for t_i in t:
            uniq_t.add(t_i)
        for p_i in p[:k]:
            if p_i in t:
                uniq_tp.add(p_i)
    return len(uniq_tp) / len(uniq_t)


def dcg_at_k(Y_true, Y_pred, k=5):
    """
    Calculate DCG at k

    :param Y_true: true labels
    :param Y_pred: ranking of predicted labels
    :param k: k
    :return: value of DCG at k
    """
    Y_true = _process_Y(Y_true)
    Y_pred = _process_Y(Y_pred)

    sum = 0
    for t, p in zip(Y_true, Y_pred):
        for i, p_i in enumerate(p[:k]):
            sum += 1 / log2(i + 2) if p_i in t else 0
    return sum / len(Y_true)


def ndcg_at_k(Y_true, Y_pred, k=5):
    """
    Calculate nDCG at k

    :param Y_true: true labels
    :param Y_pred: ranking of predicted labels
    :param k: k
    :return: value of nDCG at k
    """
    Y_true = _process_Y(Y_true)
    Y_pred = _process_Y(Y_pred)

    sum = 0
    for t, p in zip(Y_true, Y_pred):
        dcg = 0
        n = 0
        for i, p_i in enumerate(p[:k]):
            n += 1 / log2(i + 2)
            dcg += 1 / log2(i + 2) if p_i in t else 0
        sum += dcg / n
    return sum / len(Y_true)


def hamming_loss(Y_true, Y_pred):
    """
    Calculate hamming loss

    :param Y_true: true labels
    :param Y_pred: predicted labels
    :return: value of hamming loss
    """

    Y_true = _process_Y(Y_true)
    Y_pred = _process_Y(Y_pred)

    sum = 0
    for t, p in zip(Y_true, Y_pred):
        sum += len(set(t).intersection(p))

    return sum / len(Y_true)



# Copyright (c) 2020 by Marek Wydmuch
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import re
from numpy import ndarray
from scipy.sparse import csr_matrix
from ._napkinxc import CPPModel, InputDataType


class Model():
    """
    Main model class that wraps CPPModel
    """

    def __init__(self, **params):
        self._model = CPPModel() # In the feature more CPP classes may be available
        self._params = {}
        self.set_params(**params)

    def fit(self, X, Y):
        """
        Fit the model to the given training data

        :param X: training data points
        :type X: array-like, sparse matrix, list of lists of tuples (idx, value)
        :param Y: target labels
        :type Y: list of lists or tuples
        :return: None
        """
        self._model.fit(X, Y, Model._check_data_type(X), Model._check_data_type(Y))

    def fit_on_file(self, path):
        """
        Fit the model to the training data in the given file

        :param path: path to the file
        :type path: str
        :return: None
        """
        self._model.fitOnFile(path)

    def predict(self, X, top_k=5, threshold=0):
        """
        Predict labels for data points in X

        :param X: data points
        :type X: array-like, sparse matrix, list of lists of tuples (idx, value)
        :param top_k: Predict top-k labels, defaults to 5
        :type top_k: int
        :param threshold: Predict labels with probability above the threshold, defaults to 0
        :type threshold: float
        :return: list of list
        """
        return self._model.predict(X, Model._check_data_type(X), top_k, threshold)

    def predict_proba(self, X, top_k=5, threshold=0):
        """
        Predict labels with probability estimates for data points in X

        :param X: data points
        :type X: array-like, sparse matrix, list of lists of tuples (idx, value)
        :param top_k: Predict top-k labels, defaults to 5
        :type top_k: int
        :param threshold: Predict labels with probability above the threshold, defaults to 0
        :type threshold: float
        :return: list of list of tuples
        """
        return self._model.predict(X, Model._check_data_type(X), top_k, threshold)

    def predict_for_file(self, path, top_k=5, threshold=0):
        """
        Predict labels for data points in the given file

        :param path: path to the file
        :type path: str
        :param top_k: Predict top-k labels, defaults to 5
        :type top_k: int
        :param threshold: Predict labels with probability above the threshold, defaults to 0
        :type threshold: float
        :return: list of list
        """
        return self._model.predict_for_file(path, top_k, threshold)

    def predict_proba_for_file(self, path, top_k=5, threshold=0):
        """
        Predict labels with probability estimates for data points in the given file

        :param path: path to the file
        :type path: str
        :param top_k: Predict top-k labels, defaults to 5
        :type top_k: int
        :param threshold: Predict labels with probability above the threshold, defaults to 0
        :type threshold: float
        :return: list of list of tuples
        """
        return self._model.predict_proba_for_file(path, top_k, threshold)

    def get_params(self, deep=False): # deep argument for sklearn compatibility
        """
        Get parameters of this model/

        :param deep: Ignored, added for sklearn compatibility, defaults to False.
        :return: mapping of string to any
        """
        return self._params

    def set_params(self, **params):
        """
        Set parameters for this model.

        :param: \*\*params: Parameter names with their new values.
        :return: self
        """
        if 'model' in self._params and 'model' in params:
            params.pop('model')  # do not allow for changing model param

        if 'verbose' in params and params['verbose'] == True:
            params['verbose'] = 2  # override verbose

        self._params.update(params)
        params_list = []
        for k, v in params.items():
            if v is None:
                continue

            arg_k = ("--" if len(k) > 1 else "-") + Model._to_camelcase(k)
            arg_v = v
            if type(arg_v) == bool:
                arg_v = int(arg_v)
            arg_v = str(arg_v)

            params_list.extend([arg_k, arg_v])
        self._model.set_args(params_list)

        return self

    @staticmethod
    def _get_init_params(locals):
        kwargs = locals['kwargs']
        for k in ['self', '__class__', 'kwargs']:
            locals.pop(k)
        locals.update(kwargs)
        return locals

    @staticmethod
    def _to_camelcase(string):
        return re.sub(r'(?!^)_([a-zA-Z])', lambda m: m.group(1).upper(), string)

    @staticmethod
    def _check_data_type(data):
        data_type = type(data)
        if data_type == list:
            return InputDataType.list
        elif data_type == ndarray:
            return InputDataType.ndarray
        elif data_type == csr_matrix:
            return InputDataType.csr_matrix
        else:
            return -1


class PLT(Model):
    """
    Probabilistic Labels Trees (PLTs) model with linear node estimators, using CPP core.
    """

    def __init__(self,
                 output,

                 # Tree params
                 tree_type='hierarchicalKmeans',
                 arity=2,
                 max_leaves=100,
                 kmeans_eps=0.0001,
                 kmeans_balanced=True,
                 #tree_structure=None, #TODO

                 # Features params
                 hash=None,
                 features_threshold=0,
                 norm=True,
                 bias=1.0,

                 # Node classifiers params
                 optimizer="liblinear",
                 loss='log',
                 weights_threshold=0.1,
                 liblinear_c=10,
                 liblinear_eps=0.1,
                 eta=1.0,
                 epochs=1,
                 adagrad_eps=0.001,

                 # Other
                 ensemble=1,
                 seed=None,
                 threads=0,
                 verbose=0,
                 **kwargs):
        """
        Construct a Probabilistic Labels Trees model.

        :param output: directory where the model will be stored
        :type output: str
        :param tree_type: tree type to construct {'hierarchicalKmeans', 'balancedRandom', 'completeKaryRandom', 'huffman'}, defaults to 'hierarchicalKmeans'
        :type tree_type: str, optional
        :param arity: arity of tree nodes, k for k-means clustering used in hierarchical k-means tree building procedure, defaults to 2
        :type arity: int, optional
        :param max_leaves: maximum degree of pre-leaf nodes, defaults to 100
        :type max_leaves: int, optional
        :param kmeans_eps: tolerance of termination criterion of the k-means clustering used in hierarchical k-means tree building procedure, defaults to 0.0001
        :type kmeans_eps: float, optional
        :param kmeans_balanced: use balanced k-means clustering, defaults to True
        :type kmeans_balanced: bool, optional
        :param hash: hash features to a space of given size, if None or 0 disable hashing, defaults to None
        :type hash: int, optional
        :param features_threshold: prune features below given threshold, defaults to 0
        :type features_threshold: float, optional
        :param norm: unit norm feature vector, defaults to True
        :type norm: bool, optional
        :param bias: value of the bias features, defaults to 1.0
        :type bias: float, optional
        :param optimizer: optimizer used for training node classifiers {'liblinear', 'sgd', 'adagrad'}, defaults to 'libliner'
        :type optimizer: str, optional
        :param loss: loss optimized while training node classifiers {'logistic', 'l2' (squared hinge)}, defaults to 'logistic'
        :type loss: str, optional
        :param weights_threshold: threshold value for pruning weights, defaults to 0.1
        :type weights_threshold: float, optional
        :param liblinear_c: LIBLINEAR cost co-efficient, inverse regularization strength, smaller values specify stronger regularization, defaults to 10.0
        :type liblinear_c: float, optional
        :param liblinear_eps: LIBLINEAR tolerance of termination criterion, defaults to 0.1
        :type liblinear_eps: float, optional
        :param eta: step size (learning rate) for online optimizers, defaults to 1.0
        :type eta: float, optional
        :param epochs: number of training epochs for online optimizers, defaults to 1
        :type epochs: int, optional
        :param adagrad_eps: defines starting step size for AdaGrad, defaults to 0.001
        :type adagrad_eps: float, optional
        :param ensemble: number of trees in the ensemble, defaults to 1
        :type ensemble: int, optional
        :param seed: seed, if None use current system time, defaults to None
        :type seed: int, optional
        :param threads: number of threads used for training and prediction, if 0 use number of available CPUs, if -1 use number of available CPUs - 1, defaults to 0
        :type threads: int, optional
        :param verbose: if True print progress, defaults to False
        :type verbose: bool, optional
        """
        all_params = Model._get_init_params(locals())
        all_params.update({"model": "plt"})
        super(PLT, self).__init__(**all_params)


class HSM(Model):
    """
    Hierarchical Softmax model with linear node estimators, using CPP core.
    """

    def __init__(self,
                 output,

                 # Tree params
                 tree_type='hierarchicalKmeans',
                 arity=2,
                 max_leaves=100,
                 kmeans_eps=0.0001,
                 kmeans_balanced=True,
                 #tree_structure=None, #TODO

                 # Features params
                 hash=None,
                 features_threshold=0,
                 norm=True,
                 bias=1.0,

                 # Node classifiers params
                 optimizer="liblinear",
                 loss='log',
                 weights_threshold=0.1,
                 liblinear_c=10,
                 liblinear_eps=0.1,
                 eta=1.0,
                 epochs=1,
                 adagrad_eps=0.001,

                 # Other
                 ensemble=1,
                 seed=None,
                 threads=0,
                 verbose=0,
                 **kwargs):
        """
        Construct a Hierarchical Softmax model.

        :param output: directory where the model will be stored
        :type output: str
        :param tree_type: tree type to construct {'hierarchicalKmeans', 'balancedRandom', 'completeKaryRandom', 'huffman'}, defaults to 'hierarchicalKmeans'
        :type tree_type: str, optional
        :param arity: arity of tree nodes, k for k-means clustering used in hierarchical k-means tree building procedure, defaults to 2
        :type arity: int, optional
        :param max_leaves: maximum degree of pre-leaf nodes, defaults to 100
        :type max_leaves: int, optional
        :param kmeans_eps: tolerance of termination criterion of the k-means clustering used in hierarchical k-means tree building procedure, defaults to 0.0001
        :type kmeans_eps: float, optional
        :param kmeans_balanced: use balanced k-means clustering, defaults to True
        :type kmeans_balanced: bool, optional
        :param hash: hash features to a space of given size, if None or 0 disable hashing, defaults to None
        :type hash: int, optional
        :param features_threshold: prune features below given threshold, defaults to 0
        :type features_threshold: float, optional
        :param norm: unit norm feature vector, defaults to True
        :type norm: bool, optional
        :param bias: value of the bias features, defaults to 1.0
        :type bias: float, optional
        :param optimizer: optimizer used for training binary classifiers {'liblinear', 'sgd', 'adagrad'}, defaults to 'libliner'
        :type optimizer: str, optional
        :param loss: loss optimized while training binary classifiers {'logistic', 'l2' (squared hinge)}, defaults to 'logistic'
        :type loss: str, optional
        :param weights_threshold: threshold value for pruning weights, defaults to 0.1
        :type weights_threshold: float, optional
        :param liblinear_c: LIBLINEAR cost co-efficient, inverse regularization strength, smaller values specify stronger regularization, defaults to 10.0
        :type liblinear_c: float, optional
        :param liblinear_eps: LIBLINEAR tolerance of termination criterion, defaults to 0.1
        :type liblinear_eps: float, optional
        :param eta: step size (learning rate) for online optimizers, defaults to 1.0
        :type eta: float, optional
        :param epochs: number of training epochs for online optimizers, defaults to 1
        :type epochs: int, optional
        :param adagrad_eps: defines starting step size for AdaGrad, defaults to 0.001
        :type adagrad_eps: float, optional
        :param ensemble: number of trees in the ensemble, defaults to 1
        :type ensemble: int, optional
        :param seed: seed, if None use current system time, defaults to None
        :type seed: int, optional
        :param threads: number of threads used for training and prediction, if 0 use number of available CPUs, if -1 use number of available CPUs - 1, defaults to 0
        :type threads: int, optional
        :param verbose: if True print progress, defaults to False
        :type verbose: bool, optional
        """
        all_params = Model._get_init_params(locals())
        all_params.update({"model": "hsm"})
        super(HSM, self).__init__(**all_params)


class BR(Model):
    """
    Binary Relevance model with linear node estimators, using CPP core
    """

    def __init__(self,
                 output,
                 # Features params
                 hash=None,
                 features_threshold=0,
                 norm=True,
                 bias=1.0,

                 # Node classifiers params
                 optimizer="liblinear",
                 loss='log',
                 weights_threshold=0.1,
                 liblinear_c=10,
                 liblinear_eps=0.1,
                 eta=1.0,
                 epochs=1,
                 adagrad_eps=0.001,

                 # Other
                 threads=0,
                 verbose=0,
                 **kwargs):
        """
        Construct a Binary Relevance model.

        :param output: directory where the model will be stored
        :type output: str, optional
        :param hash: hash features to a space of given size, if None or 0 disable hashing, defaults to None
        :type hash: int, optional
        :param features_threshold: prune features below given threshold, defaults to 0
        :type features_threshold: float, optional
        :param norm: unit norm feature vector, defaults to True
        :type norm: bool, optional
        :param bias: value of the bias features, defaults to 1.0
        :type bias: float, optional
        :param optimizer: optimizer used for training binary classifiers {'liblinear', 'sgd', 'adagrad'}, defaults to 'libliner'
        :type optimizer: str, optional
        :param loss: loss optimized while training binary classifiers {'logistic', 'l2' (squared hinge)}, defaults to 'logistic'
        :type loss: str, optional
        :param weights_threshold: threshold value for pruning weights, defaults to 0.1
        :type weights_threshold: float, optional
        :param liblinear_c: LIBLINEAR cost co-efficient, inverse regularization strength, smaller values specify stronger regularization, defaults to 10.0
        :type liblinear_c: float, optional
        :param liblinear_eps: LIBLINEAR tolerance of termination criterion, defaults to 0.1
        :type liblinear_eps: float, optional
        :param eta: step size (learning rate) for online optimizers, defaults to 1.0
        :type eta: float, optional
        :param epochs: number of training epochs for online optimizers, defaults to 1
        :type epochs: int, optional
        :param adagrad_eps: defines starting step size for AdaGrad, defaults to 0.001
        :type adagrad_eps: float, optional
        :param threads: number of threads used for training and prediction, if 0 use number of available CPUs, if -1 use number of available CPUs - 1, defaults to 0
        :type threads: int, optional
        :param verbose: if True print progress, defaults to False
        :type verbose: bool, optional
        """
        all_params = Model._get_init_params(locals())
        all_params.update({"model": "br"})
        super(BR, self).__init__(**all_params)


class OVR(Model):
    """
    One Versus Rest model with linear node estimators, using CPP core.
    """

    def __init__(self,
                 output,
                 # Features params
                 hash=None,
                 features_threshold=0,
                 norm=True,
                 bias=1.0,

                 # Node classifiers params
                 optimizer="liblinear",
                 loss='log',
                 weights_threshold=0.1,
                 liblinear_c=10,
                 liblinear_eps=0.1,
                 eta=1.0,
                 epochs=1,
                 adagrad_eps=0.001,

                 # Other
                 threads=0,
                 verbose=0,
                 **kwargs):
        """
        Construct a One Versus Rest model.

        :param output: directory where the model will be stored
        :type output: str
        :param hash: hash features to a space of given size, if None or 0 disable hashing, defaults to None
        :type hash: int, optional
        :param features_threshold: prune features below given threshold, defaults to 0
        :type features_threshold: float, optional
        :param norm: unit norm feature vector, defaults to True
        :type norm: bool, optional
        :param bias: value of the bias features, defaults to 1.0
        :type bias: float, optional
        :param optimizer: optimizer used for training node classifiers {'liblinear', 'sgd', 'adagrad'}, defaults to 'libliner'
        :type optimizer: str, optional
        :param loss: loss optimized while training node classifiers {'logistic', 'l2' (squared hinge)}, defaults to 'logistic'
        :type loss: str, optional
        :param weights_threshold: threshold value for pruning weights, defaults to 0.1
        :type weights_threshold: float, optional
        :param liblinear_c: LIBLINEAR cost co-efficient, inverse regularization strength, smaller values specify stronger regularization, defaults to 10.0
        :type liblinear_c: float, optional
        :param liblinear_eps: LIBLINEAR tolerance of termination criterion, defaults to 0.1
        :type liblinear_eps: float, optional
        :param eta: step size (learning rate) for online optimizers, defaults to 1.0
        :type eta: float, optional
        :param epochs: number of training epochs for online optimizers, defaults to 1
        :type epochs: int, optional
        :param adagrad_eps: defines starting step size for AdaGrad, defaults to 0.001
        :type adagrad_eps: float, optional
        :param threads: number of threads used for training and prediction, if 0 use number of available CPUs, if -1 use number of available CPUs - 1, defaults to 0
        :type threads: int, optional
        :param verbose: if True print progress, defaults to False
        :type verbose: bool, optional
        """
        all_params = Model._get_init_params(locals())
        all_params.update({"model": "ovr"})
        super(OVR, self).__init__(**all_params)



import re
from cmaketools import setup
from multiprocessing import cpu_count
from os import path


def get_long_description():
    try:
        dir_path = path.dirname(path.realpath(__file__))
        with open(path.join(dir_path, "README.md"), encoding="utf-8") as readme_file:
            return readme_file.read()

    except Exception:
        raise RuntimeError("Package description retrieval failed. "
                           "Most probably something is wrong with this code and "
                           "you should create an issue at https://github.com/mwydmuch/napkinXC")


def get_version():
    try:
        dir_path = path.dirname(path.realpath(__file__))
        with open(path.join(dir_path, "CMakeLists.txt")) as cmake_file:
            lines = cmake_file.read()
            version = re.search('VERSION\s+([0-9].[0-9].[0-9]+)', lines).group(1)
            return version

    except Exception:
        raise RuntimeError("Package version retrieval failed. "
                           "Most probably something is wrong with this code and "
                           "you should create an issue at https://github.com/mwydmuch/napkinXC")


setup(
    name="napkinxc",
    version=get_version(),
    author="Marek Wydmuch",
    author_email="mwydmuch@cs.put.poznan.pl",
    description="napkinXC is an extremely simple and fast library for extreme multi-class and multi-label classification.",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    url="https://github.com/mwydmuch/napkinXC",
    keywords=['machine learning', 'extreme classification', 'multi-class classification', 'multi-label classification', 'classification'],
    classifiers=[
        'Development Status :: 4 - Beta',
        #Development Status :: 5 - Production/Stable,
        'Operating System :: POSIX :: Linux',
        'Operating System :: MacOS',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    license="MIT License",
    platforms=["Linux", "MacOS"],
    install_requires=['numpy', 'scipy', 'sklearn'],
    src_dir="python",
    packages=['napkinxc'],
    ext_module_hint=r"pybind11_add_module",
    configure_opts=['-DPYTHON=ON', '-DEXE=OFF'],
    include_package_data=True,
    #parallel=max(cpu_count(), 1) # Unknown distribution option: 'parallel'
)



#!/usr/bin/env python3

from napkinxc.datasets import load_dataset
from napkinxc.models import PLT
from napkinxc.measures import precision_at_k

# Use load_dataset function to load one of the benchmark datasets
# from XML Repository (http://manikvarma.org/downloads/XC/XMLRepository.html)
X_train, Y_train = load_dataset("eurlex-4k", "train")
X_test, Y_test = load_dataset("eurlex-4k", "test")

# Create Probabilistic Labels Tree models,
# directory "eurlex-model" will be created and used for model training and storing
plt = PLT("eurlex-model")

# Fit the model on the train dataset
plt.fit(X_train, Y_train)

# Predict only the best label for each datapoint in the test dataset
Y_pred = plt.predict(X_test, top_k=1)

# Evaluate precision at 1
print(precision_at_k(Y_test, Y_pred, k=1))
